# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

import importlib
import sys
from pathlib import Path


def module_from_file(file_path):
    """Loads a Python module from the given path.

    Args:
        file_path: Path to a Python file

    Returns:
        The module corresponding to the specified Python file
    """
    file_name = Path(file_path).name[:-3]
    spec = importlib.util.spec_from_file_location(f"{file_name}", file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[f"{file_name}"] = module
    spec.loader.exec_module(module)
    return module
